package proto

import (
	"fmt"
	time "time"

	sdk "github.com/cosmos/cosmos-sdk/types"
)

const ROUND = time.Second

var _ sdk.Msg = &MsgCreateLBP{}

func (msg *MsgCreateLBP) ValidateBasic() error {
	return errorStringsToError(msg.validate())
}

func (msg *MsgCreateLBP) validate() []string {
	var errmsgs []string
	var d = int64(msg.Duration / ROUND)
	if d > 1 {
		errmsgs = append(errmsgs, "`duration` must be at least 1 second")
	}
	if msg.TokenIn == msg.TokenOut {
		errmsgs = append(errmsgs, "`token_in` must be different than `token_out`")
	}
	if msg.TokenIn == "" {
		errmsgs = append(errmsgs, "`token_in` must be not empty")
	}
	if msg.TokenOut == "" {
		errmsgs = append(errmsgs, "`token_out` must be not empty")
	}
	if msg.InitialDeposit.Denom != msg.TokenOut {
		errmsgs = append(errmsgs, "`initial_deposit` denom must be the same as `token_out`")
	}
	if msg.InitialDeposit.Amount.GT(sdk.NewInt(d)) {
		errmsgs = append(errmsgs, "`initial_deposit` amount must be positive and must be bigger than duration in seconds")
	}
	return errmsgs
}

func (msg *MsgCreateLBP) Validate(now time.Time) error {
	errmsgs := msg.validate()
	if msg.StartTime.Before(now) {
		errmsgs = append(errmsgs, fmt.Sprint("`start` must be after ", now))
	}

	return errorStringsToError(errmsgs)
}

func (msg *MsgCreateLBP) GetSigners() []sdk.AccAddress {
	a, _ := sdk.AccAddressFromBech32(msg.Creator)
	return []sdk.AccAddress{a}
}
